#![allow(clippy::upper_case_acronyms)]

use crate::cpu::{Cpu, StatusFlag};
use crate::memory::{MemoryReader, MemoryWriter};

use anyhow::{bail, Result};

#[derive(Clone, Copy, Debug)]
pub enum Instruction {
    Valid(ValidInstruction),
    Invalid(InvalidInstruction),
}

#[derive(Clone, Copy, Debug)]
pub struct ValidInstruction {
    pub opcode: Opcode,
    pub cycles: u8,
}

#[derive(Clone, Copy, Debug)]
pub struct InvalidInstruction {
    pub opcode: u8,
}

#[derive(Clone, Copy, Debug)]
pub enum AddressingMode {
    AbsoluteA,                  // a
    AbsoluteIndexedWithX,       // a, x
    AbsoluteIndexedWithY,       // a, y
    AbsoluteIndirect,           // (a)
    AbsoluteIndexedIndirect,    // (a,x) (4), only used with the JMP instruction
    Accumulator,                // A
    Immediate,                  // #
    Implied,                    // i
    ProgramCounterRelative,     // r
    ProgramCounterRelativeTest, // r, but for instructions like BBR0-7, where the first operand
    // is a zero-page address, and the second is the relative branch
    // address
    Stack,                        // s
    ZeroPage,                     // zp
    ZeroPageIndexedWithX,         // zp, x
    ZeroPageIndexedWithY,         // zp, y
    ZeroPageIndirect,             // (zp) (4)
    ZeroPageIndexedIndirect,      // (zp, x)
    ZeroPageIndirectIndexedWithY, // (zp), y
}

// if you want to optimize this later https://llx.com/Neil/a2/opcodes.html
pub fn get_instruction(opcode: u8) -> Instruction {
    let opcodes = include!("./opcodes.in");
    opcodes[opcode as usize]
}

#[derive(Clone, Copy, Debug)]
pub enum Opcode {
    ADC(AddressingMode),
    AND(AddressingMode),
    ASL(AddressingMode),
    BBR0(AddressingMode),
    BBR1(AddressingMode),
    BBR2(AddressingMode),
    BBR3(AddressingMode),
    BBR4(AddressingMode),
    BBR5(AddressingMode),
    BBR6(AddressingMode),
    BBR7(AddressingMode),
    BBS0(AddressingMode),
    BBS1(AddressingMode),
    BBS2(AddressingMode),
    BBS3(AddressingMode),
    BBS4(AddressingMode),
    BBS5(AddressingMode),
    BBS6(AddressingMode),
    BBS7(AddressingMode),
    BCC(AddressingMode),
    BCS(AddressingMode),
    BEQ(AddressingMode),
    BIT(AddressingMode),
    BMI(AddressingMode),
    BNE(AddressingMode),
    BPL(AddressingMode),
    BRA(AddressingMode),
    BRK(AddressingMode),
    BVC(AddressingMode),
    BVS(AddressingMode),
    CLC(AddressingMode),
    CLD(AddressingMode),
    CLI(AddressingMode),
    CLV(AddressingMode),
    CMP(AddressingMode),
    CPX(AddressingMode),
    CPY(AddressingMode),
    DEC(AddressingMode),
    DEX(AddressingMode),
    DEY(AddressingMode),
    EOR(AddressingMode),
    INC(AddressingMode),
    INX(AddressingMode),
    INY(AddressingMode),
    JMP(AddressingMode),
    JSR(AddressingMode),
    LDA(AddressingMode),
    LDX(AddressingMode),
    LDY(AddressingMode),
    LSR(AddressingMode),
    NOP(AddressingMode),
    ORA(AddressingMode),
    PHA(AddressingMode),
    PHP(AddressingMode),
    PHX(AddressingMode),
    PHY(AddressingMode),
    PLA(AddressingMode),
    PLP(AddressingMode),
    PLX(AddressingMode),
    PLY(AddressingMode),
    RMB0(AddressingMode),
    RMB1(AddressingMode),
    RMB2(AddressingMode),
    RMB3(AddressingMode),
    RMB4(AddressingMode),
    RMB5(AddressingMode),
    RMB6(AddressingMode),
    RMB7(AddressingMode),
    ROL(AddressingMode),
    ROR(AddressingMode),
    RTI(AddressingMode),
    RTS(AddressingMode),
    SBC(AddressingMode),
    SEC(AddressingMode),
    SED(AddressingMode),
    SEI(AddressingMode),
    SMB0(AddressingMode),
    SMB1(AddressingMode),
    SMB2(AddressingMode),
    SMB3(AddressingMode),
    SMB4(AddressingMode),
    SMB5(AddressingMode),
    SMB6(AddressingMode),
    SMB7(AddressingMode),
    STA(AddressingMode),
    STP(AddressingMode),
    STX(AddressingMode),
    STY(AddressingMode),
    STZ(AddressingMode),
    TAX(AddressingMode),
    TAY(AddressingMode),
    TRB(AddressingMode),
    TSB(AddressingMode),
    TSX(AddressingMode),
    TXA(AddressingMode),
    TXS(AddressingMode),
    TYA(AddressingMode),
    WAI(AddressingMode),
}

#[derive(Clone, Copy)]
enum AddressingModeValue {
    Implied,              // Used when an instruction doesn't need an operation, e.g. RTS
    Relative(u8),         // Used by branch instructions to offset the program counter
    RelativeTest(u8, u8), // Used by branch instructions to offset the program counter
    Absolute(u16),        // Just a plain ol' address
    Accumulator(u8),      // This isn't technically accurate, since the datasheet says this would be
                          // Implied, but we do use the value of the accumulator as the operand so this
                          // is how i'm gonna do it
}

impl TryFrom<AddressingModeValue> for u16 {
    type Error = anyhow::Error;
    fn try_from(value: AddressingModeValue) -> Result<Self> {
        match value {
            AddressingModeValue::Implied => bail!(
                "An addressing mode value without an address was attempted to turn into a u16"
            ),
            AddressingModeValue::RelativeTest(_zero_page, offset) => Ok(offset as u16),
            AddressingModeValue::Relative(inner_value) => Ok(inner_value as u16),
            AddressingModeValue::Absolute(inner_value) => Ok(inner_value),
            AddressingModeValue::Accumulator(inner_value) => Ok(inner_value as u16),
        }
    }
}

fn get_address(mode: &AddressingMode, cpu: &mut Cpu) -> Result<AddressingModeValue> {
    fn accumulator(cpu: &mut Cpu) -> AddressingModeValue {
        let byte = cpu.a;
        AddressingModeValue::Accumulator(byte)
    }

    fn implied(_cpu: &mut Cpu) -> AddressingModeValue {
        // i
        AddressingModeValue::Implied
    }

    fn immediate(cpu: &mut Cpu) -> AddressingModeValue {
        // #
        let address: u16 = cpu.pc;
        cpu.pc = cpu.pc.wrapping_add(1);
        AddressingModeValue::Absolute(address)
    }

    fn absolute_a(cpu: &mut Cpu) -> Result<AddressingModeValue> {
        // a
        let address = cpu.read_word(cpu.pc)?;
        cpu.pc = cpu.pc.wrapping_add(2);
        Ok(AddressingModeValue::Absolute(address))
    }

    fn zero_page(cpu: &mut Cpu) -> Result<AddressingModeValue> {
        // zp
        let address = cpu.read(cpu.pc);
        cpu.pc = cpu.pc.wrapping_add(1);
        Ok(AddressingModeValue::Absolute(address as u16))
    }

    fn absolute_indexed_with_x(
        // a, y
        cpu: &mut Cpu,
    ) -> Result<AddressingModeValue> {
        let word = cpu.read_word(cpu.pc)?;
        cpu.pc = cpu.pc.wrapping_add(2);
        let address: u16 = word + cpu.x as u16;
        Ok(AddressingModeValue::Absolute(address))
    }

    fn absolute_indexed_with_y(
        // a, y
        cpu: &mut Cpu,
    ) -> Result<AddressingModeValue> {
        let word = cpu.read_word(cpu.pc)?;
        cpu.pc = cpu.pc.wrapping_add(2);
        let address: u16 = word + cpu.y as u16;
        Ok(AddressingModeValue::Absolute(address))
    }

    fn zero_page_indexed_with_x(
        // zp, x
        cpu: &mut Cpu,
    ) -> Result<AddressingModeValue> {
        let byte = cpu.read(cpu.pc);
        cpu.pc = cpu.pc.wrapping_add(1);
        let address: u16 = (byte.wrapping_add(cpu.x)) as u16;
        Ok(AddressingModeValue::Absolute(address))
    }

    fn zero_page_indexed_with_y(
        // zp, y
        cpu: &mut Cpu,
    ) -> Result<AddressingModeValue> {
        let byte = cpu.read(cpu.pc);
        cpu.pc = cpu.pc.wrapping_add(1);
        let address: u16 = (byte + cpu.y) as u16;
        Ok(AddressingModeValue::Absolute(address))
    }

    fn absolute_indirect(
        // (a)
        cpu: &mut Cpu,
    ) -> Result<AddressingModeValue> {
        let word = cpu.read_word(cpu.pc)?;
        let address = cpu.read_word(word)?;
        cpu.pc = cpu.pc.wrapping_add(2);
        Ok(AddressingModeValue::Absolute(address))
    }

    fn absolute_indexed_indirect(
        // (a, x), only used with the JMP instruction
        cpu: &mut Cpu,
    ) -> Result<AddressingModeValue> {
        let word = cpu.read_word(cpu.pc)?;
        let address = cpu.read_word(word + cpu.x as u16)?;
        cpu.pc = cpu.pc.wrapping_add(2);
        Ok(AddressingModeValue::Absolute(address))
    }

    fn zero_page_indirect(
        // (zp)
        cpu: &mut Cpu,
    ) -> Result<AddressingModeValue> {
        let byte: u8 = cpu.read(cpu.pc);
        let address: u16 = cpu.read_word(byte as u16)?;
        cpu.pc = cpu.pc.wrapping_add(1);
        Ok(AddressingModeValue::Absolute(address))
    }

    fn zero_page_indexed_indirect(
        // (zp, x)
        cpu: &mut Cpu,
    ) -> Result<AddressingModeValue> {
        let byte = cpu.read(cpu.pc);
        let address = cpu.read_word((byte.wrapping_add(cpu.x)) as u16)?; // Anytime you see something like `byte as u16`, it's using the byte as a zero-page address
        cpu.pc = cpu.pc.wrapping_add(1);
        Ok(AddressingModeValue::Absolute(address))
    }

    fn zero_page_indirect_indexed_with_y(
        // (zp), y
        cpu: &mut Cpu,
    ) -> Result<AddressingModeValue> {
        let byte: u8 = cpu.read(cpu.pc);
        let address: u16 = cpu.read_word(byte.wrapping_add(cpu.y) as u16)?;
        cpu.pc = cpu.pc.wrapping_add(1);
        Ok(AddressingModeValue::Absolute(address))
    }

    fn relative(
        // r
        cpu: &mut Cpu,
    ) -> Result<AddressingModeValue> {
        let byte: u8 = cpu.read(cpu.pc);
        cpu.pc = cpu.pc.wrapping_add(1);
        Ok(AddressingModeValue::Relative(byte))
    }

    fn relative_test(
        // r
        cpu: &mut Cpu,
    ) -> Result<AddressingModeValue> {
        let byte: u8 = cpu.read(cpu.pc);
        cpu.pc = cpu.pc.wrapping_add(1);
        let address = cpu.read(cpu.pc);
        cpu.pc = cpu.pc.wrapping_add(1);
        Ok(AddressingModeValue::RelativeTest(byte, address))
    }

    match mode {
        AddressingMode::AbsoluteA => absolute_a(cpu),
        AddressingMode::AbsoluteIndexedWithX => absolute_indexed_with_x(cpu),
        AddressingMode::AbsoluteIndexedWithY => absolute_indexed_with_y(cpu),
        AddressingMode::AbsoluteIndirect => absolute_indirect(cpu),
        AddressingMode::AbsoluteIndexedIndirect => absolute_indexed_indirect(cpu),
        AddressingMode::Immediate => Ok(immediate(cpu)),
        AddressingMode::ProgramCounterRelative => relative(cpu),
        AddressingMode::ZeroPage => zero_page(cpu),
        AddressingMode::ZeroPageIndexedWithX => zero_page_indexed_with_x(cpu),
        AddressingMode::ZeroPageIndexedWithY => zero_page_indexed_with_y(cpu),
        AddressingMode::ZeroPageIndirect => zero_page_indirect(cpu),
        AddressingMode::ZeroPageIndexedIndirect => zero_page_indexed_indirect(cpu),
        AddressingMode::ZeroPageIndirectIndexedWithY => zero_page_indirect_indexed_with_y(cpu),
        AddressingMode::Accumulator => Ok(accumulator(cpu)),
        AddressingMode::ProgramCounterRelativeTest => relative_test(cpu),
        _ => Ok(implied(cpu)),
    }
}

fn branch(cpu: &mut Cpu, condition: bool, value: AddressingModeValue) -> Result<()> {
    fn signed_byte_to_word(value: u8) -> u16 {
        let mut value = u16::from(value);
        if value & 0x80 > 0 {
            value |= 0xff00;
        }
        value
    }
    match value {
        AddressingModeValue::Relative(address) => {
            let address = signed_byte_to_word(address).wrapping_add(cpu.pc);
            if condition {
                if address & 0xff00 != cpu.pc & 0xff00 {
                    cpu.pending_cycles += 2;
                } else {
                    cpu.pending_cycles += 1;
                }
                cpu.pc = address;
            }
            Ok(())
        }
        AddressingModeValue::RelativeTest(_zero_page, address) => {
            let address = signed_byte_to_word(address).wrapping_add(cpu.pc);
            if condition {
                if address & 0xff00 != cpu.pc & 0xff00 {
                    cpu.pending_cycles += 2;
                } else {
                    cpu.pending_cycles += 1;
                }
                cpu.pc = address;
            }
            Ok(())
        }
        _ => bail!("Tried to branch but received an invalid addressing mode value"),
    }
}

impl Opcode {
    pub fn call(&self, cpu: &mut Cpu) -> Result<()> {
        match self {
            Opcode::ADC(mode) => match mode {
                AddressingMode::Immediate
                | AddressingMode::ZeroPage
                | AddressingMode::ZeroPageIndexedWithX
                | AddressingMode::AbsoluteA
                | AddressingMode::AbsoluteIndexedWithX
                | AddressingMode::AbsoluteIndexedWithY
                | AddressingMode::ZeroPageIndirect
                | AddressingMode::ZeroPageIndexedIndirect
                | AddressingMode::ZeroPageIndirectIndexedWithY => {
                    let address = get_address(mode, cpu)?;
                    let byte = cpu.read(address.try_into()?);
                    let carry = cpu.get_flag(StatusFlag::Carry);
                    let result = cpu.a as u16 + byte as u16 + carry as u16;
                    cpu.set_flag(StatusFlag::Carry, result > u16::from(u8::max_value()));
                    cpu.set_flag(StatusFlag::Zero, result as u8 == 0);
                    cpu.set_flag(
                        StatusFlag::Overflow,
                        (cpu.a ^ byte) & (cpu.a ^ result as u8) & 0b1000_0000 > 0,
                        // Truly have no idea what's happening here,
                        // check out https://docs.rs/emulator_6502/latest/src/emulator_6502/opcodes/mod.rs.html
                    );
                    cpu.set_flag(StatusFlag::Negative, cpu.is_negative(result as u8));
                    cpu.a = result as u8;
                    Ok(())
                }
                _ => bail!("An incompatible addressing mode was used for instruction {self:#?} at address {:#04x}", cpu.pc)
            },
            Opcode::AND(mode) => match mode {
                AddressingMode::Immediate
                | AddressingMode::ZeroPage
                | AddressingMode::ZeroPageIndexedWithX
                | AddressingMode::AbsoluteA
                | AddressingMode::AbsoluteIndexedWithX
                | AddressingMode::AbsoluteIndexedWithY
                | AddressingMode::ZeroPageIndirect
                | AddressingMode::ZeroPageIndexedIndirect
                | AddressingMode::ZeroPageIndirectIndexedWithY => {
                    let address = get_address(mode, cpu)?;
                    cpu.a &= cpu.read(address.try_into()?);
                    cpu.set_flag(StatusFlag::Zero, cpu.a == 0);
                    cpu.set_flag(StatusFlag::Negative, cpu.is_negative(cpu.a));
                    Ok(())
                }
                _ => bail!("An incompatible addressing mode was used for instruction {self:#?} at address {:#04x}", cpu.pc)
            },
            Opcode::ASL(mode) => {
                fn asl(cpu: &mut Cpu, value: u8) -> u8 {
                    cpu.set_flag(StatusFlag::Carry, value & 0b1000_0000 == 0b1000_0000);
                    let shifted_value = value << 1;
                    cpu.set_flag(StatusFlag::Negative, cpu.is_negative(shifted_value));
                    cpu.set_flag(StatusFlag::Zero, shifted_value == 0);
                    shifted_value
                }
                match mode {
                    AddressingMode::Accumulator => {
                        cpu.a = asl(cpu, cpu.a);
                        Ok(())
                    }
                    AddressingMode::AbsoluteA
                    | AddressingMode::AbsoluteIndexedWithX
                    | AddressingMode::ZeroPage
                    | AddressingMode::ZeroPageIndexedWithX => {
                        let address = get_address(mode, cpu)?;
                        let value = cpu.read(address.try_into()?);
                        let result = asl(cpu, value);
                        cpu.write(address.try_into()?, result);
                        Ok(())
                    }
                    _ => bail!("An incompatible addressing mode was used for instruction {self:#?} at address {:#04x}", cpu.pc)
                }
            }
            Opcode::BBR0(mode) => match mode {
                // These instructions are weird, cause they're relative, except the byte being tested is immediate,
                // so i'm not sure if what i've already written is going to model this accurately
                AddressingMode::ProgramCounterRelativeTest => match get_address(mode, cpu) {
                    Ok(value) => match value {
                        AddressingModeValue::RelativeTest(byte, _address) => {
                            let zero_page_address = byte as u16;
                            let test_byte = cpu.read(zero_page_address);
                            branch(cpu, test_byte & 0b0000_0001 != 0b0000_0001, value)?;
                            Ok(())
                        }
                        _ => bail!("An incompatible addressing mode was used for instruction {self:#?} at address {:#04x}", cpu.pc)
                    },
                    Err(error) => bail!("Tried to get address, but {error:#?}"),
                },
                _ => bail!("An incompatible addressing mode was used for instruction {self:#?} at address {:#04x}", cpu.pc)
            },
            Opcode::BBR1(mode) => match mode {
                AddressingMode::ProgramCounterRelativeTest => match get_address(mode, cpu) {
                    Ok(value) => match value {
                        AddressingModeValue::RelativeTest(byte, _address) => {
                            let zero_page_address = byte as u16;
                            let test_byte = cpu.read(zero_page_address);
                            branch(cpu, test_byte & 0b0000_0010 != 0b0000_0010, value)?;
                            Ok(())
                        }
                        _ => bail!("An incompatible addressing mode was used for instruction {self:#?} at address {:#04x}", cpu.pc)
                    },
                    Err(error) => bail!("Tried to get address, but {error:#?}"),
                },
                _ => bail!("An incompatible addressing mode was used for instruction {self:#?} at address {:#04x}", cpu.pc)
            },
            Opcode::BBR2(mode) => match mode {
                AddressingMode::ProgramCounterRelativeTest => match get_address(mode, cpu) {
                    Ok(value) => match value {
                        AddressingModeValue::RelativeTest(byte, _address) => {
                            let zero_page_address = byte as u16;
                            let test_byte = cpu.read(zero_page_address);
                            branch(cpu, test_byte & 0b0000_0100 != 0b0000_0100, value)?;
                            Ok(())
                        }
                        _ => bail!("An incompatible addressing mode was used for instruction {self:#?} at address {:#04x}", cpu.pc)
                    },
                    Err(error) => bail!("Tried to get address, but {error:#?}"),
                },
                _ => bail!("An incompatible addressing mode was used for instruction {self:#?} at address {:#04x}", cpu.pc)
            },
            Opcode::BBR3(mode) => match mode {
                AddressingMode::ProgramCounterRelativeTest => match get_address(mode, cpu) {
                    Ok(value) => match value {
                        AddressingModeValue::RelativeTest(byte, _address) => {
                            let zero_page_address = byte as u16;
                            let test_byte = cpu.read(zero_page_address);
                            branch(cpu, test_byte & 0b0000_1000 != 0b0000_1000, value)?;
                            Ok(())
                        }
                        _ => bail!("An incompatible addressing mode was used for instruction {self:#?} at address {:#04x}", cpu.pc)
                    },
                    Err(error) => bail!("Tried to get address, but {error:#?}"),
                },
                _ => bail!("An incompatible addressing mode was used for instruction {self:#?} at address {:#04x}", cpu.pc)
            },
            Opcode::BBR4(mode) => match mode {
                AddressingMode::ProgramCounterRelativeTest => match get_address(mode, cpu) {
                    Ok(value) => match value {
                        AddressingModeValue::RelativeTest(byte, _address) => {
                            let zero_page_address = byte as u16;
                            let test_byte = cpu.read(zero_page_address);
                            branch(cpu, test_byte & 0b0001_0000 != 0b0001_0000, value)?;
                            Ok(())
                        }
                        _ => bail!("An incompatible addressing mode was used for instruction {self:#?} at address {:#04x}", cpu.pc)
                    },
                    Err(error) => bail!("Tried to get address, but {error:#?}"),
                },
                _ => bail!("An incompatible addressing mode was used for instruction {self:#?} at address {:#04x}", cpu.pc)
            },
            Opcode::BBR5(mode) => match mode {
                AddressingMode::ProgramCounterRelativeTest => match get_address(mode, cpu) {
                    Ok(value) => match value {
                        AddressingModeValue::RelativeTest(byte, _address) => {
                            let zero_page_address = byte as u16;
                            let test_byte = cpu.read(zero_page_address);
                            branch(cpu, test_byte & 0b0010_0000 != 0b0010_0000, value)?;
                            Ok(())
                        }
                        _ => bail!("An incompatible addressing mode was used for instruction {self:#?} at address {:#04x}", cpu.pc)
                    },
                    Err(error) => bail!("Tried to get address, but {error:#?}"),
                },
                _ => bail!("An incompatible addressing mode was used for instruction {self:#?} at address {:#04x}", cpu.pc)
            },
            Opcode::BBR6(mode) => match mode {
                AddressingMode::ProgramCounterRelativeTest => match get_address(mode, cpu) {
                    Ok(value) => match value {
                        AddressingModeValue::RelativeTest(byte, _address) => {
                            let zero_page_address = byte as u16;
                            let test_byte = cpu.read(zero_page_address);
                            branch(cpu, test_byte & 0b0100_0000 != 0b0100_0000, value)?;
                            Ok(())
                        }
                        _ => bail!("An incompatible addressing mode was used for instruction {self:#?} at address {:#04x}", cpu.pc)
                    },
                    Err(error) => bail!("Tried to get address, but {error:#?}"),
                },
                _ => bail!("An incompatible addressing mode was used for instruction {self:#?} at address {:#04x}", cpu.pc)
            },
            Opcode::BBR7(mode) => match mode {
                AddressingMode::ProgramCounterRelativeTest => match get_address(mode, cpu) {
                    Ok(value) => match value {
                        AddressingModeValue::RelativeTest(byte, _address) => {
                            let zero_page_address = byte as u16;
                            let test_byte = cpu.read(zero_page_address);
                            branch(cpu, test_byte & 0b1000_0000 != 0b1000_0000, value)?;
                            Ok(())
                        }
                        _ => bail!("An incompatible addressing mode was used for instruction {self:#?} at address {:#04x}", cpu.pc)
                    },
                    Err(error) => bail!("Tried to get address, but {error:#?}"),
                },
                _ => bail!("An incompatible addressing mode was used for instruction {self:#?} at address {:#04x}", cpu.pc)
            },
            Opcode::BBS0(mode) => match mode {
                AddressingMode::ProgramCounterRelativeTest => match get_address(mode, cpu) {
                    Ok(value) => match value {
                        AddressingModeValue::RelativeTest(byte, _address) => {
                            let zero_page_address = byte as u16;
                            let test_byte = cpu.read(zero_page_address);
                            branch(cpu, test_byte & 0b0000_0001 == 0b0000_0001, value)?;
                            Ok(())
                        }
                        _ => bail!("An incompatible addressing mode was used for instruction {self:#?} at address {:#04x}", cpu.pc)
                    },
                    Err(error) => bail!("Tried to get address, but {error:#?}"),
                },
                _ => bail!("An incompatible addressing mode was used for instruction {self:#?} at address {:#04x}", cpu.pc)
            },
            Opcode::BBS1(mode) => match mode {
                AddressingMode::ProgramCounterRelativeTest => match get_address(mode, cpu) {
                    Ok(value) => match value {
                        AddressingModeValue::RelativeTest(byte, _address) => {
                            let zero_page_address = byte as u16;
                            let test_byte = cpu.read(zero_page_address);
                            branch(cpu, test_byte & 0b0000_0010 == 0b0000_0010, value)?;
                            Ok(())
                        }
                        _ => bail!("An incompatible addressing mode was used for instruction {self:#?} at address {:#04x}", cpu.pc)
                    },
                    Err(error) => bail!("Tried to get address, but {error:#?}"),
                },
                _ => bail!("An incompatible addressing mode was used for instruction {self:#?} at address {:#04x}", cpu.pc)
            },
            Opcode::BBS2(mode) => match mode {
                AddressingMode::ProgramCounterRelativeTest => match get_address(mode, cpu) {
                    Ok(value) => match value {
                        AddressingModeValue::RelativeTest(byte, _address) => {
                            let zero_page_address = byte as u16;
                            let test_byte = cpu.read(zero_page_address);
                            branch(cpu, test_byte & 0b0000_0100 == 0b0000_0100, value)?;
                            Ok(())
                        }
                        _ => bail!("An incompatible addressing mode was used for instruction {self:#?} at address {:#04x}", cpu.pc)
                    },
                    Err(error) => bail!("Tried to get address, but {error:#?}"),
                },
                _ => bail!("An incompatible addressing mode was used for instruction {self:#?} at address {:#04x}", cpu.pc)
            },
            Opcode::BBS3(mode) => match mode {
                AddressingMode::ProgramCounterRelativeTest => match get_address(mode, cpu) {
                    Ok(value) => match value {
                        AddressingModeValue::RelativeTest(byte, _address) => {
                            let zero_page_address = byte as u16;
                            let test_byte = cpu.read(zero_page_address);
                            branch(cpu, test_byte & 0b0000_1000 == 0b0000_1000, value)?;
                            Ok(())
                        }
                        _ => bail!("An incompatible addressing mode was used for instruction {self:#?} at address {:#04x}", cpu.pc)
                    },
                    Err(error) => bail!("Tried to get address, but {error:#?}"),
                },
                _ => bail!("An incompatible addressing mode was used for instruction {self:#?} at address {:#04x}", cpu.pc)
            },
            Opcode::BBS4(mode) => match mode {
                AddressingMode::ProgramCounterRelativeTest => match get_address(mode, cpu) {
                    Ok(value) => match value {
                        AddressingModeValue::RelativeTest(byte, _address) => {
                            let zero_page_address = byte as u16;
                            let test_byte = cpu.read(zero_page_address);
                            branch(cpu, test_byte & 0b0001_0000 == 0b0001_0000, value)?;
                            Ok(())
                        }
                        _ => bail!("An incompatible addressing mode was used for instruction {self:#?} at address {:#04x}", cpu.pc)
                    },
                    Err(error) => bail!("Tried to get address, but {error:#?}"),
                },
                _ => bail!("An incompatible addressing mode was used for instruction {self:#?} at address {:#04x}", cpu.pc)
            },
            Opcode::BBS5(mode) => match mode {
                AddressingMode::ProgramCounterRelativeTest => match get_address(mode, cpu) {
                    Ok(value) => match value {
                        AddressingModeValue::RelativeTest(byte, _address) => {
                            let zero_page_address = byte as u16;
                            let test_byte = cpu.read(zero_page_address);
                            branch(cpu, test_byte & 0b0010_0000 == 0b0010_0000, value)?;
                            Ok(())
                        }
                        _ => bail!("An incompatible addressing mode was used for instruction {self:#?} at address {:#04x}", cpu.pc)
                    },
                    Err(error) => bail!("Tried to get address, but {error:#?}"),
                },
                _ => bail!("An incompatible addressing mode was used for instruction {self:#?} at address {:#04x}", cpu.pc)
            },
            Opcode::BBS6(mode) => match mode {
                AddressingMode::ProgramCounterRelativeTest => match get_address(mode, cpu) {
                    Ok(value) => match value {
                        AddressingModeValue::RelativeTest(byte, _address) => {
                            let zero_page_address = byte as u16;
                            let test_byte = cpu.read(zero_page_address);
                            branch(cpu, test_byte & 0b0100_0000 == 0b0100_0000, value)?;
                            Ok(())
                        }
                        _ => bail!("An incompatible addressing mode was used for instruction {self:#?} at address {:#04x}", cpu.pc)
                    },
                    Err(error) => bail!("Tried to get address, but {error:#?}"),
                },
                _ => bail!("An incompatible addressing mode was used for instruction {self:#?} at address {:#04x}", cpu.pc)
            },
            Opcode::BBS7(mode) => match mode {
                AddressingMode::ProgramCounterRelativeTest => match get_address(mode, cpu) {
                    Ok(value) => match value {
                        AddressingModeValue::RelativeTest(byte, _address) => {
                            let zero_page_address = byte as u16;
                            let test_byte = cpu.read(zero_page_address);
                            branch(cpu, test_byte & 0b1000_0000 == 0b1000_0000, value)?;
                            Ok(())
                        }
                        _ => bail!("An incompatible addressing mode was used for instruction {self:#?} at address {:#04x}", cpu.pc)
                    },
                    Err(error) => bail!("Tried to get address, but {error:#?}"),
                },
                _ => bail!("An incompatible addressing mode was used for instruction {self:#?} at address {:#04x}", cpu.pc)
            },
            Opcode::BCC(mode) => match mode {
                AddressingMode::ProgramCounterRelative => {
                    let address = get_address(mode, cpu)?;
                    branch(cpu, !cpu.get_flag(StatusFlag::Carry), address)?;
                    Ok(())
                }
                _ => bail!("An incompatible addressing mode was used for instruction {self:#?} at address {:#04x}", cpu.pc)
            },
            Opcode::BCS(mode) => match mode {
                AddressingMode::ProgramCounterRelative => {
                    let address = get_address(mode, cpu)?;
                    branch(cpu, cpu.get_flag(StatusFlag::Carry), address)?;
                    Ok(())
                }
                _ => bail!("An incompatible addressing mode was used for instruction {self:#?} at address {:#04x}", cpu.pc)
            },
            Opcode::BEQ(mode) => match mode {
                AddressingMode::ProgramCounterRelative => {
                    let address = get_address(mode, cpu)?;
                    branch(cpu, cpu.get_flag(StatusFlag::Zero), address)?;
                    Ok(())
                }
                _ => bail!("An incompatible addressing mode was used for instruction {self:#?} at address {:#04x}", cpu.pc)
            },
            Opcode::BIT(mode) => match mode {
                AddressingMode::Immediate
                | AddressingMode::ZeroPage
                | AddressingMode::ZeroPageIndexedWithX
                | AddressingMode::AbsoluteA
                | AddressingMode::AbsoluteIndexedWithX => {
                    let address = get_address(mode, cpu)?;
                    let result = cpu.a & cpu.read(address.try_into()?);
                    cpu.set_flag(StatusFlag::Zero, result == 0);
                    cpu.set_flag(
                        StatusFlag::Overflow,
                        result & StatusFlag::Overflow as u8 == StatusFlag::Overflow as u8,
                    );
                    cpu.set_flag(
                        StatusFlag::Negative,
                        result & StatusFlag::Negative as u8 == StatusFlag::Negative as u8,
                    );
                    Ok(())
                }
                _ => bail!("An incompatible addressing mode was used for instruction {self:#?} at address {:#04x}", cpu.pc)
            },
            Opcode::BMI(mode) => match mode {
                AddressingMode::ProgramCounterRelative => {
                    let address = get_address(mode, cpu)?;
                    branch(cpu, cpu.get_flag(StatusFlag::Negative), address)?;
                    Ok(())
                }
                _ => bail!("An incompatible addressing mode was used for instruction {self:#?} at address {:#04x}", cpu.pc)
            },
            Opcode::BNE(mode) => match mode {
                AddressingMode::ProgramCounterRelative => {
                    let address = get_address(mode, cpu)?;
                    branch(cpu, !cpu.get_flag(StatusFlag::Zero), address)?;
                    Ok(())
                }
                _ => bail!("An incompatible addressing mode was used for instruction {self:#?} at address {:#04x}", cpu.pc)
            },
            Opcode::BPL(mode) => match mode {
                AddressingMode::ProgramCounterRelative => {
                    let address = get_address(mode, cpu)?;
                    branch(cpu, !cpu.get_flag(StatusFlag::Negative), address)?;
                    Ok(())
                }
                _ => bail!("An incompatible addressing mode was used for instruction {self:#?} at address {:#04x}", cpu.pc)
            },
            Opcode::BRA(mode) => match mode {
                AddressingMode::ProgramCounterRelative => {
                    let address = get_address(mode, cpu)?;
                    branch(cpu, true, address)?;
                    Ok(())
                }
                _ => bail!("An incompatible addressing mode was used for instruction {self:#?} at address {:#04x}", cpu.pc)
            },
            Opcode::BRK(mode) => match mode {
                AddressingMode::Implied => {
                    cpu.set_flag(StatusFlag::Brk, true);
                    Ok(())
                    //panic!("Interrupts unimplemented!");
                }
                _ => bail!("An incompatible addressing mode was used for instruction {self:#?} at address {:#04x}", cpu.pc)
            },
            Opcode::BVC(mode) => match mode {
                AddressingMode::ProgramCounterRelative => {
                    let address = get_address(mode, cpu)?;
                    branch(cpu, !cpu.get_flag(StatusFlag::Overflow), address)?;
                    Ok(())
                }
                _ => bail!("An incompatible addressing mode was used for instruction {self:#?} at address {:#04x}", cpu.pc)
            },
            Opcode::BVS(mode) => match mode {
                AddressingMode::ProgramCounterRelative => {
                    let address = get_address(mode, cpu)?;
                    branch(cpu, cpu.get_flag(StatusFlag::Overflow), address)?;
                    Ok(())
                }
                _ => bail!("An incompatible addressing mode was used for instruction {self:#?} at address {:#04x}", cpu.pc)
            },
            Opcode::CLC(mode) => match mode {
                AddressingMode::Implied => {
                    cpu.set_flag(StatusFlag::Carry, false);
                    Ok(())
                }
                _ => bail!("An incompatible addressing mode was used for instruction {self:#?} at address {:#04x}", cpu.pc)
            },
            Opcode::CLD(mode) => match mode {
                AddressingMode::Implied => {
                    cpu.set_flag(StatusFlag::Decimal, false);
                    Ok(())
                }
                _ => bail!("An incompatible addressing mode was used for instruction {self:#?} at address {:#04x}", cpu.pc)
            },
            Opcode::CLI(mode) => match mode {
                AddressingMode::Implied => {
                    cpu.set_flag(StatusFlag::IrqDisable, false);
                    Ok(())
                }
                _ => bail!("An incompatible addressing mode was used for instruction {self:#?} at address {:#04x}", cpu.pc)
            },
            Opcode::CLV(mode) => match mode {
                AddressingMode::Implied => {
                    cpu.set_flag(StatusFlag::Overflow, false);
                    Ok(())
                }
                _ => bail!("An incompatible addressing mode was used for instruction {self:#?} at address {:#04x}", cpu.pc)
            },
            Opcode::CMP(mode) => match mode {
                AddressingMode::Immediate
                | AddressingMode::ZeroPage
                | AddressingMode::ZeroPageIndexedWithX
                | AddressingMode::AbsoluteA
                | AddressingMode::AbsoluteIndexedWithX
                | AddressingMode::AbsoluteIndexedWithY
                | AddressingMode::ZeroPageIndirect
                | AddressingMode::ZeroPageIndexedIndirect
                | AddressingMode::ZeroPageIndirectIndexedWithY => {
                    let address = get_address(mode, cpu)?;
                    let byte = cpu.read(address.try_into()?);
                    cpu.set_flag(StatusFlag::Carry, cpu.a >= byte);
                    cpu.set_flag(StatusFlag::Zero, cpu.a == byte);
                    cpu.set_flag(StatusFlag::Negative, cpu.a <= byte);
                    Ok(())
                }
                _ => bail!("An incompatible addressing mode was used for instruction {self:#?} at address {:#04x}", cpu.pc)
            },
            Opcode::CPX(mode) => match mode {
                AddressingMode::Immediate
                | AddressingMode::ZeroPage
                | AddressingMode::AbsoluteA => {
                    let address = get_address(mode, cpu)?;
                    let byte = cpu.read(address.try_into()?);
                    cpu.set_flag(StatusFlag::Carry, cpu.x >= byte);
                    cpu.set_flag(StatusFlag::Zero, cpu.x == byte);
                    cpu.set_flag(StatusFlag::Negative, cpu.is_negative(cpu.x - byte));
                    Ok(())
                }
                _ => bail!("An incompatible addressing mode was used for instruction {self:#?} at address {:#04x}", cpu.pc)
            },
            Opcode::CPY(mode) => match mode {
                AddressingMode::Immediate
                | AddressingMode::ZeroPage
                | AddressingMode::AbsoluteA => {
                    let address = get_address(mode, cpu)?;
                    let byte = cpu.read(address.try_into()?);
                    cpu.set_flag(StatusFlag::Carry, cpu.y >= byte);
                    cpu.set_flag(StatusFlag::Zero, cpu.y == byte);
                    cpu.set_flag(StatusFlag::Negative, cpu.y <= byte);
                    Ok(())
                }
                _ => bail!("An incompatible addressing mode was used for instruction {self:#?} at address {:#04x}", cpu.pc)
            },
            Opcode::DEC(mode) => match mode {
                AddressingMode::ZeroPage
                | AddressingMode::ZeroPageIndexedWithX
                | AddressingMode::AbsoluteA
                | AddressingMode::AbsoluteIndexedWithX => {
                    let address = get_address(mode, cpu)?;
                    let byte = cpu.read(address.try_into()?);
                    let dec_byte = byte.wrapping_sub(1);
                    cpu.write(address.try_into()?, dec_byte);
                    cpu.set_flag(StatusFlag::Zero, dec_byte == 0);
                    cpu.set_flag(StatusFlag::Negative, cpu.is_negative(dec_byte));
                    Ok(())
                }
                _ => bail!("An incompatible addressing mode was used for instruction {self:#?} at address {:#04x}", cpu.pc)
            },
            Opcode::DEX(mode) => match mode {
                AddressingMode::Implied => {
                    cpu.x = cpu.x.wrapping_sub(1);
                    cpu.set_flag(StatusFlag::Zero, cpu.x == 0);
                    cpu.set_flag(StatusFlag::Negative, cpu.is_negative(cpu.x));
                    Ok(())
                }
                _ => bail!("An incompatible addressing mode was used for instruction {self:#?} at address {:#04x}", cpu.pc)
            },
            Opcode::DEY(mode) => match mode {
                AddressingMode::Implied => {
                    cpu.y = cpu.y.wrapping_sub(1);
                    cpu.set_flag(StatusFlag::Zero, cpu.y == 0);
                    cpu.set_flag(StatusFlag::Negative, cpu.is_negative(cpu.y));
                    Ok(())
                }
                _ => bail!("An incompatible addressing mode was used for instruction {self:#?} at address {:#04x}", cpu.pc)
            },
            Opcode::EOR(mode) => match mode {
                AddressingMode::Immediate
                | AddressingMode::ZeroPage
                | AddressingMode::ZeroPageIndexedWithX
                | AddressingMode::AbsoluteA
                | AddressingMode::AbsoluteIndexedWithX
                | AddressingMode::AbsoluteIndexedWithY
                | AddressingMode::ZeroPageIndirect
                | AddressingMode::ZeroPageIndexedIndirect
                | AddressingMode::ZeroPageIndirectIndexedWithY => {
                    let address = get_address(mode, cpu)?;
                    cpu.a ^= cpu.read(address.try_into()?);
                    cpu.set_flag(StatusFlag::Zero, cpu.a == 0);
                    cpu.set_flag(StatusFlag::Negative, cpu.is_negative(cpu.a));
                    Ok(())
                }
                _ => bail!("An incompatible addressing mode was used for instruction {self:#?} at address {:#04x}", cpu.pc)
            },
            Opcode::INC(mode) => match mode {
                AddressingMode::Accumulator => {
                    cpu.a = cpu.a.wrapping_add(1);
                    cpu.set_flag(StatusFlag::Zero, cpu.a == 0);
                    cpu.set_flag(StatusFlag::Negative, cpu.is_negative(cpu.a));
                    Ok(())
                }
                AddressingMode::ZeroPage
                | AddressingMode::ZeroPageIndexedWithX
                | AddressingMode::AbsoluteA
                | AddressingMode::AbsoluteIndexedWithX => {
                    let address = get_address(mode, cpu)?;
                    let byte = cpu.read(address.try_into()?);
                    cpu.set_flag(StatusFlag::Zero, byte.wrapping_add(1) == 0);
                    cpu.set_flag(StatusFlag::Negative, cpu.is_negative(byte.wrapping_add(1)));
                    cpu.write(address.try_into()?, byte.wrapping_add(1));
                    Ok(())
                }
                _ => bail!("An incompatible addressing mode was used for instruction {self:#?} at address {:#04x}", cpu.pc)
            },
            Opcode::INX(mode) => match mode {
                AddressingMode::Implied => {
                    cpu.x = cpu.x.wrapping_add(1);
                    cpu.set_flag(StatusFlag::Zero, cpu.x == 0);
                    cpu.set_flag(StatusFlag::Negative, cpu.is_negative(cpu.x));
                    Ok(())
                }
                _ => bail!("An incompatible addressing mode was used for instruction {self:#?} at address {:#04x}", cpu.pc)
            },
            Opcode::INY(mode) => match mode {
                AddressingMode::Implied => {
                    cpu.y = cpu.y.wrapping_add(1);
                    cpu.set_flag(StatusFlag::Zero, cpu.y == 0);
                    cpu.set_flag(StatusFlag::Negative, cpu.is_negative(cpu.y));
                    Ok(())
                }
                _ => bail!("An incompatible addressing mode was used for instruction {self:#?} at address {:#04x}", cpu.pc)
            },
            Opcode::JMP(mode) => match mode {
                AddressingMode::AbsoluteA
                | AddressingMode::AbsoluteIndirect
                | AddressingMode::AbsoluteIndexedIndirect => {
                    let address = get_address(mode, cpu)?;
                    cpu.pc = address.try_into()?;
                    Ok(())
                }
                _ => bail!("An incompatible addressing mode was used for instruction {self:#?} at address {:#04x}", cpu.pc)
            },
            Opcode::JSR(mode) => match mode {
                AddressingMode::AbsoluteA => {
                    let return_address = cpu.pc - 1;
                    cpu.push_stack_word(return_address)?;
                    let address = get_address(mode, cpu)?;
                    cpu.pc = address.try_into()?;
                    Ok(())
                }
                _ => bail!("An incompatible addressing mode was used for instruction {self:#?} at address {:#04x}", cpu.pc)
            },
            Opcode::LDA(mode) => match mode {
                AddressingMode::AbsoluteA
                | AddressingMode::AbsoluteIndexedWithX
                | AddressingMode::AbsoluteIndexedWithY
                | AddressingMode::Immediate
                | AddressingMode::ZeroPage
                | AddressingMode::ZeroPageIndirect
                | AddressingMode::ZeroPageIndexedIndirect
                | AddressingMode::ZeroPageIndexedWithX
                | AddressingMode::ZeroPageIndirectIndexedWithY => {
                    let address = get_address(mode, cpu)?;
                    let byte = cpu.read(address.try_into()?);
                    cpu.a = byte;
                    cpu.set_flag(StatusFlag::Negative, cpu.is_negative(cpu.a));
                    cpu.set_flag(StatusFlag::Zero, cpu.a == 0);
                    Ok(())
                }
                _ => bail!("An incompatible addressing mode was used for instruction {self:#?} at address {:#04x}", cpu.pc)
            },
            Opcode::LDX(mode) => match mode {
                AddressingMode::AbsoluteA
                | AddressingMode::AbsoluteIndexedWithY
                | AddressingMode::Immediate
                | AddressingMode::ZeroPage
                | AddressingMode::ZeroPageIndexedWithY => {
                    let address = get_address(mode, cpu)?;
                    let byte = cpu.read(address.try_into()?);
                    cpu.x = byte;
                    cpu.set_flag(StatusFlag::Negative, cpu.is_negative(cpu.x));
                    cpu.set_flag(StatusFlag::Zero, cpu.x == 0);
                    Ok(())
                }
                _ => bail!("An incompatible addressing mode was used for instruction {self:#?} at address {:#04x}", cpu.pc)
            },
            Opcode::LDY(mode) => match mode {
                AddressingMode::AbsoluteA
                | AddressingMode::AbsoluteIndexedWithX
                | AddressingMode::Immediate
                | AddressingMode::ZeroPage
                | AddressingMode::ZeroPageIndexedWithX => {
                    let address = get_address(mode, cpu)?;
                    let byte = cpu.read(address.try_into()?);
                    cpu.y = byte;
                    cpu.set_flag(StatusFlag::Negative, cpu.is_negative(cpu.y));
                    cpu.set_flag(StatusFlag::Zero, cpu.y == 0);
                    Ok(())
                }
                _ => bail!("An incompatible addressing mode was used for instruction {self:#?} at address {:#04x}", cpu.pc)
            },
            Opcode::LSR(mode) => {
                fn lsr(cpu: &mut Cpu, value: u8) -> u8 {
                    cpu.set_flag(StatusFlag::Carry, value & 1 == 1);
                    let shifted_value = value >> 1;
                    cpu.set_flag(StatusFlag::Zero, shifted_value == 0);
                    cpu.set_flag(StatusFlag::Negative, cpu.is_negative(value));
                    shifted_value
                }
                match mode {
                    AddressingMode::Accumulator => {
                        cpu.a = lsr(cpu, cpu.a);
                        Ok(())
                    }
                    AddressingMode::AbsoluteA
                    | AddressingMode::AbsoluteIndexedWithX
                    | AddressingMode::ZeroPage
                    | AddressingMode::ZeroPageIndexedWithX => {
                        let address = get_address(mode, cpu)?;
                        let value = cpu.read(address.try_into()?);
                        let result = lsr(cpu, value);
                        cpu.write(address.try_into()?, result);
                        Ok(())
                    }
                    _ => bail!("An incompatible addressing mode was used for instruction {self:#?} at address {:#04x}", cpu.pc)
                }
            }
            Opcode::NOP(_mode) => {
                cpu.pc = cpu.pc.wrapping_add(1);
                Ok(())
            }
            Opcode::ORA(mode) => match mode {
                AddressingMode::Immediate
                | AddressingMode::ZeroPage
                | AddressingMode::ZeroPageIndexedWithX
                | AddressingMode::AbsoluteA
                | AddressingMode::AbsoluteIndexedWithX
                | AddressingMode::AbsoluteIndexedWithY
                | AddressingMode::ZeroPageIndirect
                | AddressingMode::ZeroPageIndexedIndirect
                | AddressingMode::ZeroPageIndirectIndexedWithY => {
                    let address = get_address(mode, cpu)?;
                    cpu.a |= cpu.read(address.try_into()?);
                    cpu.set_flag(StatusFlag::Zero, cpu.a == 0);
                    cpu.set_flag(StatusFlag::Negative, cpu.is_negative(cpu.a));
                    Ok(())
                }
                _ => bail!("An incompatible addressing mode was used for instruction {self:#?} at address {:#04x}", cpu.pc)
            },
            Opcode::PHA(mode) => match mode {
                AddressingMode::Stack => {
                    cpu.push_stack(cpu.a)?;
                    Ok(())
                }
                _ => bail!("An incompatible addressing mode was used for instruction {self:#?} at address {:#04x}", cpu.pc)
            },
            Opcode::PHP(mode) => match mode {
                AddressingMode::Stack => {
                    cpu.push_stack(cpu.p)?;
                    Ok(())
                }
                _ => bail!("An incompatible addressing mode was used for instruction {self:#?} at address {:#04x}", cpu.pc)
            },
            Opcode::PHX(mode) => match mode {
                AddressingMode::Stack => {
                    cpu.push_stack(cpu.x)?;
                    Ok(())
                }
                _ => bail!("An incompatible addressing mode was used for instruction {self:#?} at address {:#04x}", cpu.pc)
            },
            Opcode::PHY(mode) => match mode {
                AddressingMode::Stack => {
                    cpu.push_stack(cpu.y)?;
                    Ok(())
                }
                _ => bail!("An incompatible addressing mode was used for instruction {self:#?} at address {:#04x}", cpu.pc)
            },
            Opcode::PLA(mode) => match mode {
                AddressingMode::Stack => {
                    cpu.a = cpu.pop_stack()?;
                    Ok(())
                }
                _ => bail!("An incompatible addressing mode was used for instruction {self:#?} at address {:#04x}", cpu.pc)
            },
            Opcode::PLP(mode) => match mode {
                AddressingMode::Stack => {
                    cpu.p = cpu.pop_stack()?;
                    Ok(())
                }
                _ => bail!("An incompatible addressing mode was used for instruction {self:#?} at address {:#04x}", cpu.pc)
            },
            Opcode::PLX(mode) => match mode {
                AddressingMode::Stack => {
                    cpu.x = cpu.pop_stack()?;
                    Ok(())
                }
                _ => bail!("An incompatible addressing mode was used for instruction {self:#?} at address {:#04x}", cpu.pc)
            },
            Opcode::PLY(mode) => match mode {
                AddressingMode::Stack => {
                    cpu.y = cpu.pop_stack()?;
                    Ok(())
                }
                _ => bail!("An incompatible addressing mode was used for instruction {self:#?} at address {:#04x}", cpu.pc)
            },
            Opcode::RMB0(mode) => match mode {
                AddressingMode::ZeroPage => {
                    let address = get_address(mode, cpu)?;
                    let byte = cpu.read(address.try_into()?);
                    let reset_byte = byte & 0b1111_1110;
                    cpu.write(address.try_into()?, reset_byte);
                    Ok(())
                }
                _ => bail!("An incompatible addressing mode was used for instruction {self:#?} at address {:#04x}", cpu.pc)
            },
            Opcode::RMB1(mode) => match mode {
                AddressingMode::ZeroPage => {
                    let address = get_address(mode, cpu)?;
                    let byte = cpu.read(address.try_into()?);
                    let reset_byte = byte & 0b1111_1101;
                    cpu.write(address.try_into()?, reset_byte);
                    Ok(())
                }
                _ => bail!("An incompatible addressing mode was used for instruction {self:#?} at address {:#04x}", cpu.pc)
            },
            Opcode::RMB2(mode) => match mode {
                AddressingMode::ZeroPage => {
                    let address = get_address(mode, cpu)?;
                    let byte = cpu.read(address.try_into()?);
                    let reset_byte = byte & 0b1111_1011;
                    cpu.write(address.try_into()?, reset_byte);
                    Ok(())
                }
                _ => bail!("An incompatible addressing mode was used for instruction {self:#?} at address {:#04x}", cpu.pc)
            },
            Opcode::RMB3(mode) => match mode {
                AddressingMode::ZeroPage => {
                    let address = get_address(mode, cpu)?;
                    let byte = cpu.read(address.try_into()?);
                    let reset_byte = byte & 0b1111_0111;
                    cpu.write(address.try_into()?, reset_byte);
                    Ok(())
                }
                _ => bail!("An incompatible addressing mode was used for instruction {self:#?} at address {:#04x}", cpu.pc)
            },
            Opcode::RMB4(mode) => match mode {
                AddressingMode::ZeroPage => {
                    let address = get_address(mode, cpu)?;
                    let byte = cpu.read(address.try_into()?);
                    let reset_byte = byte & 0b1110_1111;
                    cpu.write(address.try_into()?, reset_byte);
                    Ok(())
                }
                _ => bail!("An incompatible addressing mode was used for instruction {self:#?} at address {:#04x}", cpu.pc)
            },
            Opcode::RMB5(mode) => match mode {
                AddressingMode::ZeroPage => {
                    let address = get_address(mode, cpu)?;
                    let byte = cpu.read(address.try_into()?);
                    let reset_byte = byte & 0b1101_1111;
                    cpu.write(address.try_into()?, reset_byte);
                    Ok(())
                }
                _ => bail!("An incompatible addressing mode was used for instruction {self:#?} at address {:#04x}", cpu.pc)
            },
            Opcode::RMB6(mode) => match mode {
                AddressingMode::ZeroPage => {
                    let address = get_address(mode, cpu)?;
                    let byte = cpu.read(address.try_into()?);
                    let reset_byte = byte & 0b1011_1111;
                    cpu.write(address.try_into()?, reset_byte);
                    Ok(())
                }
                _ => bail!("An incompatible addressing mode was used for instruction {self:#?} at address {:#04x}", cpu.pc)
            },
            Opcode::RMB7(mode) => match mode {
                AddressingMode::ZeroPage => {
                    let address = get_address(mode, cpu)?;
                    let byte = cpu.read(address.try_into()?);
                    let reset_byte = byte & 0b0111_1111;
                    cpu.write(address.try_into()?, reset_byte);
                    Ok(())
                }
                _ => bail!("An incompatible addressing mode was used for instruction {self:#?} at address {:#04x}", cpu.pc)
            },
            Opcode::ROL(mode) => {
                fn rol(cpu: &mut Cpu, value: u8) -> u8 {
                    let carry = cpu.get_flag(StatusFlag::Carry) as u8;
                    cpu.set_flag(StatusFlag::Carry, value >> 7 == 1);
                    let shifted_value = (value << 1) + carry;
                    cpu.set_flag(StatusFlag::Negative, cpu.is_negative(shifted_value));
                    cpu.set_flag(StatusFlag::Zero, shifted_value == 0);
                    shifted_value
                }
                match mode {
                    AddressingMode::Accumulator => {
                        cpu.a = rol(cpu, cpu.a);
                        Ok(())
                    }
                    AddressingMode::AbsoluteA
                    | AddressingMode::AbsoluteIndexedWithX
                    | AddressingMode::ZeroPage
                    | AddressingMode::ZeroPageIndexedWithX => {
                        let address = get_address(mode, cpu)?;
                        let value = cpu.read(address.try_into()?);
                        let result = rol(cpu, value);
                        cpu.write(address.try_into()?, result);
                        Ok(())
                    }
                    _ => bail!("An incompatible addressing mode was used for instruction {self:#?} at address {:#04x}", cpu.pc)
                }
            }
            Opcode::ROR(mode) => {
                fn ror(cpu: &mut Cpu, value: u8) -> u8 {
                    let carry = cpu.get_flag(StatusFlag::Carry) as u8;
                    cpu.set_flag(StatusFlag::Carry, value & 1 == 1);
                    let shifted_value = (value >> 1) + (carry << 7);
                    cpu.set_flag(StatusFlag::Negative, cpu.is_negative(shifted_value));
                    cpu.set_flag(StatusFlag::Zero, shifted_value == 0);
                    shifted_value
                }
                match mode {
                    AddressingMode::Accumulator => {
                        cpu.a = ror(cpu, cpu.a);
                        Ok(())
                    }
                    AddressingMode::AbsoluteA
                    | AddressingMode::AbsoluteIndexedWithX
                    | AddressingMode::ZeroPage
                    | AddressingMode::ZeroPageIndexedWithX => {
                        let address = get_address(mode, cpu)?;
                        let value = cpu.read(address.try_into()?);
                        let result = ror(cpu, value);
                        cpu.write(address.try_into()?, result);
                        Ok(())
                    }
                    _ => bail!("An incompatible addressing mode was used for instruction {self:#?} at address {:#04x}", cpu.pc)
                }
            }
            Opcode::RTI(mode) => match mode {
                AddressingMode::Implied => {
                    cpu.p = cpu.pop_stack()?;
                    cpu.pc = cpu.pop_stack_word()?;
                    Ok(())
                }
                _ => bail!("An incompatible addressing mode was used for instruction {self:#?} at address {:#04x}", cpu.pc)
            },
            Opcode::RTS(mode) => match mode {
                AddressingMode::Stack => {
                    let return_address = cpu.pop_stack_word()?;
                    cpu.pc = return_address + 3; // Go back to where we jsr'ed, skipping the operand
                    Ok(())
                }
                _ => bail!("An incompatible addressing mode was used for instruction {self:#?} at address {:#04x}", cpu.pc)
            },
            Opcode::SBC(mode) => match mode {
                AddressingMode::Immediate
                | AddressingMode::ZeroPage
                | AddressingMode::ZeroPageIndexedWithX
                | AddressingMode::AbsoluteA
                | AddressingMode::AbsoluteIndexedWithX
                | AddressingMode::AbsoluteIndexedWithY
                | AddressingMode::ZeroPageIndirect
                | AddressingMode::ZeroPageIndexedIndirect
                | AddressingMode::ZeroPageIndirectIndexedWithY => {
                    let address = get_address(mode, cpu)?;
                    let byte = cpu.read(address.try_into()?);
                    let carry = cpu.get_flag(StatusFlag::Carry);
                    let result = cpu.a as u16 + byte as u16 + !carry as u16;
                    cpu.set_flag(StatusFlag::Carry, result > u16::from(u8::max_value()));
                    cpu.set_flag(StatusFlag::Zero, result as u8 == 0);
                    cpu.set_flag(
                        StatusFlag::Overflow,
                        (cpu.a ^ byte) & (cpu.a ^ result as u8) & 0b1000_0000 > 0,
                        // Truly have no idea what's happening here,
                        // check out https://docs.rs/emulator_6502/latest/src/emulator_6502/opcodes/mod.rs.html
                    );
                    cpu.set_flag(StatusFlag::Negative, cpu.is_negative(result as u8));
                    Ok(())
                }
                _ => bail!("An incompatible addressing mode was used for instruction {self:#?} at address {:#04x}", cpu.pc)
            },
            Opcode::SEC(mode) => match mode {
                AddressingMode::Implied => {
                    cpu.set_flag(StatusFlag::Carry, true);
                    Ok(())
                }
                _ => bail!("An incompatible addressing mode was used for instruction {self:#?} at address {:#04x}", cpu.pc)
            },
            Opcode::SED(mode) => match mode {
                AddressingMode::Implied => {
                    cpu.set_flag(StatusFlag::Decimal, true);
                    Ok(())
                }
                _ => bail!("An incompatible addressing mode was used for instruction {self:#?} at address {:#04x}", cpu.pc)
            },
            Opcode::SEI(mode) => match mode {
                AddressingMode::Implied => {
                    cpu.set_flag(StatusFlag::IrqDisable, true);
                    Ok(())
                }
                _ => bail!("An incompatible addressing mode was used for instruction {self:#?} at address {:#04x}", cpu.pc)
            },
            Opcode::SMB0(mode) => match mode {
                AddressingMode::ZeroPage => {
                    let address = get_address(mode, cpu)?;
                    let byte = cpu.read(address.try_into()?);
                    let reset_byte = byte | 0b0000_0001;
                    cpu.write(address.try_into()?, reset_byte);
                    Ok(())
                }
                _ => bail!("An incompatible addressing mode was used for instruction {self:#?} at address {:#04x}", cpu.pc)
            },
            Opcode::SMB1(mode) => match mode {
                AddressingMode::ZeroPage => {
                    let address = get_address(mode, cpu)?;
                    let byte = cpu.read(address.try_into()?);
                    let reset_byte = byte | 0b0000_0010;
                    cpu.write(address.try_into()?, reset_byte);
                    Ok(())
                }
                _ => bail!("An incompatible addressing mode was used for instruction {self:#?} at address {:#04x}", cpu.pc)
            },
            Opcode::SMB2(mode) => match mode {
                AddressingMode::ZeroPage => {
                    let address = get_address(mode, cpu)?;
                    let byte = cpu.read(address.try_into()?);
                    let reset_byte = byte | 0b0000_0100;
                    cpu.write(address.try_into()?, reset_byte);
                    Ok(())
                }
                _ => bail!("An incompatible addressing mode was used for instruction {self:#?} at address {:#04x}", cpu.pc)
            },
            Opcode::SMB3(mode) => match mode {
                AddressingMode::ZeroPage => {
                    let address = get_address(mode, cpu)?;
                    let byte = cpu.read(address.try_into()?);
                    let reset_byte = byte | 0b0000_1000;
                    cpu.write(address.try_into()?, reset_byte);
                    Ok(())
                }
                _ => bail!("An incompatible addressing mode was used for instruction {self:#?} at address {:#04x}", cpu.pc)
            },
            Opcode::SMB4(mode) => match mode {
                AddressingMode::ZeroPage => {
                    let address = get_address(mode, cpu)?;
                    let byte = cpu.read(address.try_into()?);
                    let reset_byte = byte | 0b0001_0000;
                    cpu.write(address.try_into()?, reset_byte);
                    Ok(())
                }
                _ => bail!("An incompatible addressing mode was used for instruction {self:#?} at address {:#04x}", cpu.pc)
            },
            Opcode::SMB5(mode) => match mode {
                AddressingMode::ZeroPage => {
                    let address = get_address(mode, cpu)?;
                    let byte = cpu.read(address.try_into()?);
                    let reset_byte = byte | 0b0010_0000;
                    cpu.write(address.try_into()?, reset_byte);
                    Ok(())
                }
                _ => bail!("An incompatible addressing mode was used for instruction {self:#?} at address {:#04x}", cpu.pc)
            },
            Opcode::SMB6(mode) => match mode {
                AddressingMode::ZeroPage => {
                    let address = get_address(mode, cpu)?;
                    let byte = cpu.read(address.try_into()?);
                    let reset_byte = byte & 0b0100_0000;
                    cpu.write(address.try_into()?, reset_byte);
                    Ok(())
                }
                _ => bail!("An incompatible addressing mode was used for instruction {self:#?} at address {:#04x}", cpu.pc)
            },
            Opcode::SMB7(mode) => match mode {
                AddressingMode::ZeroPage => {
                    let address = get_address(mode, cpu)?;
                    let byte = cpu.read(address.try_into()?);
                    let reset_byte = byte | 0b1000_0000;
                    cpu.write(address.try_into()?, reset_byte);
                    Ok(())
                }
                _ => bail!("An incompatible addressing mode was used for instruction {self:#?} at address {:#04x}", cpu.pc)
            },
            Opcode::STA(mode) => match mode {
                AddressingMode::AbsoluteA
                | AddressingMode::AbsoluteIndexedWithX
                | AddressingMode::AbsoluteIndexedWithY
                | AddressingMode::ZeroPage
                | AddressingMode::ZeroPageIndirect
                | AddressingMode::ZeroPageIndexedIndirect
                | AddressingMode::ZeroPageIndexedWithX
                | AddressingMode::ZeroPageIndirectIndexedWithY => {
                    let address = get_address(mode, cpu)?;
                    cpu.write(address.try_into()?, cpu.a);
                    Ok(())
                }
                _ => bail!("An incompatible addressing mode was used for instruction {self:#?} at address {:#04x}", cpu.pc)
            },
            Opcode::STP(mode) => match mode {
                AddressingMode::Implied => {
                    cpu.stop();
                    Ok(())
                }
                _ => bail!("An incompatible addressing mode was used for instruction {self:#?} at address {:#04x}", cpu.pc)
            },
            Opcode::STX(mode) => match mode {
                AddressingMode::AbsoluteA
                | AddressingMode::ZeroPage
                | AddressingMode::ZeroPageIndexedWithY => {
                    let address = get_address(mode, cpu)?;
                    cpu.write(address.try_into()?, cpu.x);
                    Ok(())
                }
                _ => bail!("An incompatible addressing mode was used for instruction {self:#?} at address {:#04x}", cpu.pc)
            },
            Opcode::STY(mode) => match mode {
                AddressingMode::AbsoluteA
                | AddressingMode::ZeroPage
                | AddressingMode::ZeroPageIndexedWithX => {
                    let address = get_address(mode, cpu)?;
                    cpu.write(address.try_into()?, cpu.y);
                    Ok(())
                }
                _ => bail!("An incompatible addressing mode was used for instruction {self:#?} at address {:#04x}", cpu.pc)
            },
            Opcode::STZ(mode) => match mode {
                AddressingMode::AbsoluteA
                | AddressingMode::AbsoluteIndexedWithX
                | AddressingMode::ZeroPage
                | AddressingMode::ZeroPageIndexedWithX => {
                    let address = get_address(mode, cpu)?;
                    cpu.write(address.try_into()?, 0);
                    Ok(())
                }
                _ => bail!("An incompatible addressing mode was used for instruction {self:#?} at address {:#04x}", cpu.pc)
            },
            Opcode::TAX(mode) => match mode {
                AddressingMode::Implied => {
                    cpu.x = cpu.a;
                    cpu.set_flag(StatusFlag::Negative, cpu.is_negative(cpu.x));
                    cpu.set_flag(StatusFlag::Zero, cpu.x == 0);
                    Ok(())
                }
                _ => bail!("An incompatible addressing mode was used for instruction {self:#?} at address {:#04x}", cpu.pc)
            },
            Opcode::TAY(mode) => match mode {
                AddressingMode::Implied => {
                    cpu.y = cpu.a;
                    cpu.set_flag(StatusFlag::Negative, cpu.is_negative(cpu.y));
                    cpu.set_flag(StatusFlag::Zero, cpu.y == 0);
                    Ok(())
                }
                _ => bail!("An incompatible addressing mode was used for instruction {self:#?} at address {:#04x}", cpu.pc)
            },
            Opcode::TRB(mode) => match mode {
                // Still not really sure when you would
                // use TRB or TSB, but this was helpful
                // (Garth rocks) http://forum.6502.org/viewtopic.php?t=48
                AddressingMode::AbsoluteA | AddressingMode::ZeroPage => {
                    let address = get_address(mode, cpu)?;
                    let byte = cpu.read(address.try_into()?);
                    cpu.write(address.try_into()?, !cpu.a & byte);
                    cpu.set_flag(StatusFlag::Zero, cpu.a & byte > 0);
                    Ok(())
                }
                _ => bail!("An incompatible addressing mode was used for instruction {self:#?} at address {:#04x}", cpu.pc)
            },
            Opcode::TSB(mode) => match mode {
                AddressingMode::AbsoluteA | AddressingMode::ZeroPage => {
                    let address = get_address(mode, cpu)?;
                    let byte = cpu.read(address.try_into()?);
                    cpu.write(address.try_into()?, cpu.a | byte);
                    Ok(())
                }
                _ => bail!("An incompatible addressing mode was used for instruction {self:#?} at address {:#04x}", cpu.pc)
            },
            Opcode::TSX(mode) => match mode {
                AddressingMode::Implied => {
                    cpu.x = cpu.s;
                    cpu.set_flag(StatusFlag::Negative, cpu.is_negative(cpu.x));
                    cpu.set_flag(StatusFlag::Zero, cpu.x == 0);
                    Ok(())
                }
                _ => bail!("An incompatible addressing mode was used for instruction {self:#?} at address {:#04x}", cpu.pc)
            },
            Opcode::TXA(mode) => match mode {
                AddressingMode::Implied => {
                    cpu.a = cpu.x;
                    cpu.set_flag(StatusFlag::Negative, cpu.is_negative(cpu.a));
                    cpu.set_flag(StatusFlag::Zero, cpu.a == 0);
                    Ok(())
                }
                _ => bail!("An incompatible addressing mode was used for instruction {self:#?} at address {:#04x}", cpu.pc)
            },
            Opcode::TXS(mode) => match mode {
                AddressingMode::Implied => {
                    cpu.s = cpu.x;
                    Ok(())
                }
                _ => bail!("An incompatible addressing mode was used for instruction {self:#?} at address {:#04x}", cpu.pc)
            },
            Opcode::TYA(mode) => match mode {
                AddressingMode::Implied => {
                    cpu.a = cpu.y;
                    cpu.set_flag(StatusFlag::Negative, cpu.is_negative(cpu.a));
                    cpu.set_flag(StatusFlag::Zero, cpu.a == 0);
                    Ok(())
                }
                _ => bail!("An incompatible addressing mode was used for instruction {self:#?} at address {:#04x}", cpu.pc)
            },
            Opcode::WAI(mode) => match mode {
                AddressingMode::Implied => {
                    cpu.wait_for_interrupt();
                    Ok(())
                }
                _ => bail!("An incompatible addressing mode was used for instruction {self:#?} at address {:#04x}", cpu.pc)
            },
        }
    }
}
