// use std::time::Duration;

mod tabs;
mod term;
use std::{
    io::Result,
    sync::{mpsc::Receiver, Arc, Mutex},
    time::Duration,
};

use crossterm::event::{self, poll, Event, KeyCode, KeyEvent, KeyEventKind};
use ratatui::{
    layout::Layout,
    prelude::*,
    widgets::{Block, Cell, Paragraph, Row, Table, TableState},
};

use crate::{
    cpu::{CpuController, CpuState},
    memory::{MemHandle, MemoryReader, MemoryWriter},
};

pub struct App {
    cpu: CpuController,
    state: Receiver<CpuState>,
    memory: MemHandle,
    running: bool,
    table_state: TableState,
}

impl MemoryReader for App {
    fn read(&self, address: u16) -> u8 {
        self.memory.read(address)
    }
}
impl MemoryWriter for App {
    fn write(&self, address: u16, data: u8) {
        self.memory.write(address, data)
    }
}

impl App {
    pub fn new(cpu: CpuController, memory: MemHandle, state: Receiver<CpuState>) -> Self {
        Self {
            cpu,
            memory,
            running: true,
            table_state: TableState::default(),
            state,
        }
    }
    pub fn update(&mut self, terminal: &mut Terminal<impl Backend>) -> Result<()> {
        self.draw(terminal)?;
        self.handle_events()?;
        Ok(())
    }

    /// Draw a single frame of the app.
    fn draw(&self, terminal: &mut Terminal<impl Backend>) -> Result<()> {
        terminal
            .draw(|frame| {
                frame.render_widget(self, frame.size());
            })
            .unwrap();
        Ok(())
    }

    fn handle_events(&mut self) -> Result<()> {
        if poll(Duration::from_secs_f32(1.0 / 25.0))? {
            match event::read()? {
                Event::Key(key) if key.kind == KeyEventKind::Press => self.handle_key_press(key),
                _ => {}
            }
        }
        Ok(())
    }

    fn handle_key_press(&mut self, key: KeyEvent) {
        // let mut cpu = self.cpu.lock().unwrap();
        match key.code {
            KeyCode::Enter => {
                if !self.running {
                    // cpu.cycle()
                }
            }
            KeyCode::Char(' ') => {
                self.running = !self.running;
                match self.running {
                    true => self.cpu.stop(),
                    false => self.cpu.resume(),
                }
            }
            _ => {}
        };
    }
}

/// Implement Widget for &App rather than for App as we would otherwise have to clone or copy the
/// entire app state on every frame. For this example, the app state is small enough that it doesn't
/// matter, but for larger apps this can be a significant performance improvement.
impl Widget for &App {
    fn render(self, area: Rect, buf: &mut Buffer) {
        self.cpu.data();
        let [cpu_area, memory_area] =
            Layout::horizontal([Constraint::Percentage(50), Constraint::Fill(1)]).areas(area);
        let [memory_table, memory_info] =
            Layout::vertical([Constraint::Fill(1), Constraint::Percentage(10)]).areas(memory_area);

        let cpu = self.state.recv().unwrap();
        let cpu_info = format!("a: {a:#04x}, x: {x:#04x}, y: {y:#04x}, pc: {pc:#06x}, sp: {s:#04x}, sr: {p:#010b}, irq: {irq:?}, nmi: {nmi:?}", a = cpu.a, x = cpu.x, y = cpu.y, pc = cpu.pc, s = cpu.s, p = cpu.p, irq = cpu.irq, nmi = cpu.nmi);

        Paragraph::new(cpu_info)
            .block(Block::bordered().title("cpu info!"))
            .render(cpu_area, buf);

        // let table_height = memory_table.rows().count() - 2;
        // TODO: impl Iterator for MemHandle so we can get references to the underlying data from other threads without cloning
        // let rows: Vec<Row> = self.memory.data()[0..table_height * 16]
        //     .chunks(16)
        //     .map(|chunk| {
        //         chunk
        //             .iter()
        //             .map(|content| Cell::from(Text::from(format!("{content:x}"))))
        //             .collect::<Row>()
        //     })
        //     .collect();
        let zero = self.memory.read(0);
        let rows: Vec<Row> = vec![Row::new([format!("{zero}")])];
        let widths = vec![Constraint::Length(2); 16];
        Widget::render(
            Table::new(rows, widths)
                .header(
                    Row::new(vec![
                        "0", "1", "2", "3", "4", "5", "6", "7", "8", "9", "a", "b", "c", "d", "e",
                        "f",
                    ])
                    .style(Style::new().bold()), // To add space between the header and the rest of the rows, specify the margin
                                                 // .bottom_margin(1),
                )
                .block(Block::bordered().title_top("memory")),
            memory_table,
            buf,
        );
        Paragraph::new(format!("program counter: {:#04x}", cpu.pc))
            .block(Block::bordered().title_top("info"))
            .render(memory_info, buf);
    }
}

// impl Widget for App {
//     fn render_title_bar(&self, area: Rect, buf: &mut Buffer) {
//         // let layout = Layout::horizontal([Constraint::Min(0), Constraint::Length(43)]);
//         // let [title] = layout.areas(area);

//         Span::styled("Ratatui", Style::default()).render(area, buf);
//         // let titles = Tab::iter().map(Tab::title);
//         // Tabs::new(titles)
//         //     .style(THEME.tabs)
//         //     .highlight_style(THEME.tabs_selected)
//         //     .select(self.tab as usize)
//         //     .divider("")
//         //     .padding("", "")
//         //     .render(tabs, buf);
//     }

//     // fn render_selected_tab(&self, area: Rect, buf: &mut Buffer) {
//     //     match self.tab {
//     //         Tab::About => self.about_tab.render(area, buf),
//     //         Tab::Recipe => self.recipe_tab.render(area, buf),
//     //         Tab::Email => self.email_tab.render(area, buf),
//     //         Tab::Traceroute => self.traceroute_tab.render(area, buf),
//     //         Tab::Weather => self.weather_tab.render(area, buf),
//     //     };
//     // }
// }
