use crossterm::{
    cursor::{MoveTo, RestorePosition, SavePosition},
    execute, queue,
    style::{Color, PrintStyledContent, Stylize},
};

use crate::{
    cpu::CpuController,
    memory::{MemHandle, MemoryReader},
    types::{Byte, Word},
};
use std::{
    fs::File,
    io::{Read, StdoutLock, Write},
    path::Path,
    process::exit,
    sync::mpsc::Sender,
    time::Instant,
};

const FG_COLOR: u32 = 0xFFCC00;
const BG_COLOR: u32 = 0x110500;
// // BGR for softbuffer
// const FG_COLOR: u32 = 0x00CCFF;
// const BG_COLOR: u32 = 0x000511;
//
pub fn get_char_bin<P>(char_rom: Option<P>) -> Vec<u8>
where
    P: AsRef<Path>,
{
    match char_rom {
        Some(path) => {
            let mut file = File::open(path).unwrap();
            let mut bin = vec![0; 0x8000];
            file.read_exact(&mut bin).unwrap();
            // println!("reading char rom");
            bin
        }
        None => include_bytes!("./roms/cozette.rom").to_vec(),
    }
}

trait Renderer {
    fn render(&mut self) {}
}

pub struct WindowRenderer {
    char_rom: Vec<u8>,
    window: Sender<Vec<u32>>,
    memory: MemHandle,
}

impl WindowRenderer {
    pub fn new<P>(memory: MemHandle, char_rom: Option<P>, window: Sender<Vec<u32>>) -> Self
    where
        P: AsRef<Path>,
    {
        let char_rom = get_char_bin(char_rom);
        Self {
            memory,
            window,
            char_rom,
        }
    }
}

impl Renderer for WindowRenderer {
    fn render(&mut self) {
        // the rest of this function is arcane wizardry
        // based on the specifics of george's weird
        // display and characters... don't fuck around w it
        let mut i = 0;
        let mut buffer = vec![0; 512 * 380];
        for char_row in 0..29 {
            for char_col in 0..64 {
                let ascii = self.read(0x6000 + i);
                i += 1;
                for row in 0..13 {
                    let byte = self.char_rom[ascii as usize + (row * 0x100)];
                    for bit_index in (0..8).rev() {
                        let buffer_index =
                            ((char_row) * 13 + (row)) * 512 + (char_col * 8 + bit_index);
                        if (byte << bit_index) & 0x80 == 0x80 {
                            buffer[buffer_index] = FG_COLOR;
                        } else {
                            buffer[buffer_index] = BG_COLOR;
                        }
                    }
                }
            }

            let _ = self.window.send(buffer.clone());
        }
    }
}

impl MemoryReader for WindowRenderer {
    fn read(&self, address: Word) -> Byte {
        self.memory.read(address)
    }
}

pub struct TerminalRenderer<'a> {
    memory: MemHandle,
    // stdout: Stdout,
    stdout: StdoutLock<'a>,
}

impl<'a> TerminalRenderer<'a> {
    pub fn new(memory: MemHandle, stdout: StdoutLock<'a>) -> Self {
        Self { memory, stdout }
    }
}

impl MemoryReader for TerminalRenderer<'_> {
    fn read(&self, address: Word) -> Byte {
        self.memory.read(address)
    }
}

const ASCII_LOOPUP: [&str; 256] = [
    " ", "░", "▒", "▓", "♡", "♥", "⭐", "✭", "", "✦", "✨", "♀", "♂", "⚢", "⚣", "⚥", "♩", "♪",
    "♫", "♬", "ﱝ", "", "", "", "奄", "奔", "婢", "ﱜ", "ﱛ", "", "", "", " ", "!", "\"", "#",
    "$", "%", "&", "\'", "(", ")", "*", "+", ",", "-", ".", "/", "0", "1", "2", "3", "4", "5", "6",
    "7", "8", "9", ":", ";", "<", "=", ">", "?", "@", "A", "B", "C", "D", "E", "F", "G", "H", "I",
    "J", "K", "L", "M", "N", "O", "P", "Q", "R", "S", "T", "U", "V", "W", "X", "Y", "Z", "[", "\\",
    "]", "^", "_", "`", "a", "b", "c", "d", "e", "f", "g", "h", "i", "j", "k", "l", "m", "n", "o",
    "p", "q", "r", "s", "t", "u", "v", "w", "x", "y", "z", "{", "|", "}", "~", "─", "│", "┌", "└",
    "├", "┤", "┬", "┴", "┼", "╭", "╮", "╯", "╰", "╱", "╲", "╳", "═", "║", "╔", "╗", "╚", "╝", "╠",
    "╣", "╦", "╩", "╬", "", "", "", "", "", "", "", "", "", "", "", "", "", "▁", "▂",
    "▃", "▄", "▅", "▆", "▇", "█", "▉", "▊", "▋", "▌", "▍", "▎", "▏", "ʕ", "·", "ᴥ", "ʔ", "▖", "▗",
    "▘", "▙", "▚", "▛", "▜", "▝", "▞", "▟", "←", "↑", "→", "↓", "⭠", "⭡", "⭢", "⭣", "⮀", "⮁", "⮂",
    "⮃", "", "", "⏳", "", "", "", "", "", "", "", "", "", "", "", "🔮", "", "",
    "◇", "◈", "🌑", "🌒", "🌓", "🌔", "🌕", "🌖", "🌗", "🌘", "", "✔", "✘", "◆", "", "", "",
    "", "", "", "", "🎁", "", "", "", "", "⚐", "⚑", "", "", "",
];

impl Renderer for TerminalRenderer<'_> {
    fn render(&mut self) {
        // let now = Instant::now();
        let _ = execute!(self.stdout, SavePosition);
        let mut i = 0;
        for char_row in 0..29 {
            for char_col in 0..64 {
                let ascii = self.read(0x6000 + i);
                i += 1;
                let char = ASCII_LOOPUP[ascii as usize];
                let _ = queue!(
                    // FG_COLOR = 0xFFCC00
                    // BG_COLOR = 0x110500
                    self.stdout,
                    MoveTo(char_col, char_row),
                    PrintStyledContent(
                        char.with(Color::Rgb {
                            r: 0xFF,
                            g: 0xCC,
                            b: 0x00
                        })
                        .on(Color::Rgb {
                            r: 0x11,
                            g: 0x05,
                            b: 0x00
                        })
                    )
                );
            }
        }
        let _ = self.stdout.flush();
        let _ = execute!(self.stdout, RestorePosition);
        // let elapsed = now.elapsed();
        // println!("{elapsed:?}");
        // exit(0);
    }
}

pub struct Screen<'a> {
    // renderer: Box<dyn Renderer>,
    renderer: TerminalRenderer<'a>,
    controller: CpuController,
}

impl<'a> Screen<'a> {
    // pub fn new(cpu_controller: CpuController, renderer: Box<dyn Renderer>) -> Self {
    pub fn new(renderer: TerminalRenderer<'a>, controller: CpuController) -> Self {
        Self {
            renderer,
            controller,
        }
    }
    pub fn draw(&mut self) {
        self.controller.irq();
        self.renderer.render();
    }

    // pub fn run(&mut self) {
    //     loop {
    //         // sleep(Duration::from_millis(16));
    //         self.draw();
    //     }
    // }
}
